; hacker port firmware
; version 0.3 date 2004-05-24
; part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
; Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

	TITLE		"USB procedures for hackerport firmware"

	include		"P18F442.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG
	extern		SEQ_NO_RECV, SEQ_NO_SEND



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		DEBUG_LOG



;===========================================================================================
					; constants
;===========================================================================================

; addresses of usb chip registers (USBN9604)
USB_MCNTRL	equ	h'00'
USB_CCONF	equ	h'01'
USB_FAR		equ	h'04'
USB_NFSR	equ	h'05'
USB_ALTEV	equ	h'08'
USB_NAKEV	equ	h'0E'
USB_NAKMSK	equ	h'0F'
USB_EPC0	equ	h'20'
USB_TXD0	equ	h'21'
USB_TXS0	equ	h'22'
USB_TXC0	equ	h'23'
USB_RXD0	equ	h'25'
USB_RXS0	equ	h'26'
USB_RXC0	equ	h'27'
USB_EPC1	equ	h'28'
USB_TXD1	equ	h'29'
USB_TXS1	equ	h'2A'
USB_TXC1	equ	h'2B'
USB_EPC2	equ	h'2C'
USB_RXD1	equ	h'2D'
USB_RXS1	equ	h'2E'
USB_RXC1	equ	h'2F'
USB_EPC3	equ	h'30'
USB_TXD2	equ	h'31'
USB_TXS2	equ	h'32'
USB_TXC2	equ	h'33'
USB_EPC4	equ	h'34'
USB_RXD2	equ	h'35'
USB_RXS2	equ	h'36'
USB_RXC2	equ	h'37'
USB_EPC5	equ	h'38'
USB_TXD3	equ	h'39'
USB_TXS3	equ	h'3A'
USB_TXC3	equ	h'3B'
USB_EPC6	equ	h'3C'
USB_RXD3	equ	h'3D'
USB_RXS3	equ	h'3E'
USB_RXC3	equ	h'3F'



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

USB_BITS0	res		1		; bits for usb
USB_BITS1	res		1
#define		USB_IS_PRE_INIT	USB_BITS0,0	; if usb is pre-initialized (3.3 volts stabilizing)
#define		USB_IS_INIT	USB_BITS0,1	; if usb is initialized
#define		USB_S0_SEND	USB_BITS0,2	; currently sending a message on endpoint 0
#define		USB_S0_PACKET	USB_BITS0,3	; currently sending a packet of a message on endpoint 0
#define		USB_0_TOGGLE	USB_BITS0,4	; toggle bit of endpoint 0
						; - use this toggle bit for next packet to send
						; - gained from last received toggle bit by inverting it
#define		USB_CONF	USB_BITS0,5	; if we are configured
#define		USB_S1_TOGGLE	USB_BITS0,6	; toggle bit of endpoint 1
						; - use this toggle bit for next packet to send
#define		USB_S1_ACT	USB_BITS0,7	; currently sending on endpoint 1
#define		USB_S2_TOGGLE	USB_BITS1,0	; toggle bit of endpoint 2
						; - use this toggle bit for next packet to send
#define		USB_S2_ACT	USB_BITS1,1	; currently sending on endpoint 2
#define		USB_S3_TOGGLE	USB_BITS1,2	; toggle bit of endpoint 3
						; - use this toggle bit for next packet to send
#define		USB_S3_ACT	USB_BITS1,3	; currently sending on endpoint 3

USB_R0_CNT	res		1		; counter for received bytes of EP0

USB_S0_LEN	res		1		; length of rest of message to send
USB_S0_LEN_OLD	res		1		; backup of above (needed when sending a packet again)
USB_S0_PTR	res		3		; pointer to rest of message to send (little endian)
USB_S0_PTR_OLD	res		3		; backup of above (needed when sending a packet again)

; buffer to receive messages from EP0 in
USB_R0_BUF	res		0		; subdivided into 2 1-byte regs and 3 2-byte regs
USB_BM_REQ_TYPE	res		1		; bmRequestType
USB_B_REQUEST	res		1		; bRequest
USB_W_VALUE	res		2		; wValue (little endian)
USB_W_INDEX	res		2		; wIndex (little endian)
USB_W_LENGTH	res		2		; wLength (little endian)

; counter for number of endpoint to use for sending next time
USB_Sx_EP	res		1

; length of buffer to receive messages on data endpoints
; - set to > 0 when a message is received
; - set to 0 when message is processed
USB_R1_CNT	res		1		; number of received bytes on endpoint 1
USB_R2_CNT	res		1		; number of received bytes on endpoint 2
USB_R3_CNT	res		1		; number of received bytes on endpoint 3

; length of buffer to send messages on data endpoints
; - set to > 0 when a message shall be sent
; - set to 0 when message is sent (put into USB chip's FIFO)
; (the sequence number is prepended automatically,
;  so the maximum data length (i.e. maximum value of this register) is 63)
USB_Sx_CNT	res		1		; number of data bytes to send on endpoint 1, 2 or 3



;===========================================================================================
ram_usb		UDATA				; normal RAM for usb
;===========================================================================================

; buffer to receive messages on data endpoint
USB_R1_BUF	res		64		; receive buffer for endpoint 1
USB_R2_BUF	res		64		; receive buffer for endpoint 2
USB_R3_BUF	res		64		; receive buffer for endpoint 3

; buffer to send messages on data endpoint
; (the sequence number is not contained in this buffer,
;  so the maximum data length is 63)
USB_Sx_BUF	res		63		; send data buffer for endpoint 1, 2 or 3



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; usb initialization
;-------------------------------------------------------------------------------------------
USB_INIT:
; reset usb chip
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nRST		; set reset
	bsf	PIN_USB_nCS		; clear chip select
	bsf	PIN_USB_nRST		; clear reset
; initialize variables
	bcf	USB_IS_PRE_INIT		; usb not yet pre-initialized
	bcf	USB_IS_INIT		; usb not yet initialized
	bcf	BIT_USB_PWR_OK		; power may not be taken from usb
	clrf	USB_R1_CNT		; no packet received yet
	clrf	USB_R2_CNT
	clrf	USB_R3_CNT
	clrf	USB_Sx_CNT		; no packet to send yet
; done
	return

;-------------------------------------------------------------------------------------------
; read a usb chip register
; input:	W		register address (0..63)
; output:	A1_REG		value read from register
; changes:	-
;-------------------------------------------------------------------------------------------
USB_READ_REG:
; disable interrupts
;!!!
; read register
	bsf	PIN_USB_A0		; output address select bit (A0=1)
	movwf	PORT_BUS_O		; write register address to data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	setf	PORT_BUS_T		; set data bus port back to input
	bcf	PIN_USB_A0		; output data select bit (A0=0)
	bcf	PIN_USB_nRD		; set read
	movff	PORT_BUS_I,A1_REG	; read data
	bsf	PIN_USB_nRD		; clear read
	bsf	PIN_USB_nCS		; clear chip select
; enable interrupts
;!!!
; done
	return

;-------------------------------------------------------------------------------------------
; read a usb chip register multiple times
; input:	W		register address (0..63)
;		FSR0		pointer to buffer for data
;		A1_REG		number of bytes (must be > 0)
; changes:	A1_REG, FSR0
;-------------------------------------------------------------------------------------------
USB_READ_REG_EX:
; disable interrupts
;!!!
; read register multiple times
	bsf	PIN_USB_A0		; output address select bit (A0=1)
	movwf	PORT_BUS_O		; write register address to data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	setf	PORT_BUS_T		; set data bus port back to input
	bcf	PIN_USB_A0		; output data select bit (A0=0)
USB_READ_REG_EX_LOOP:
	bcf	PIN_USB_nRD		; set read
	movff	PORT_BUS_I,POSTINC0	; read data
	bsf	PIN_USB_nRD		; clear read
	decfsz	A1_REG			; loop
	bra	USB_READ_REG_EX_LOOP
	bsf	PIN_USB_nCS		; clear chip select
; enable interrupts
;!!!
; done
	return

;-------------------------------------------------------------------------------------------
; write a usb chip register
; input:	W		register address (0..63)
;		A1_REG		value to write to register
; changes:	-
;-------------------------------------------------------------------------------------------
USB_WRITE_REG:
; disable interrupts
;!!!
; write register
	bsf	PIN_USB_A0		; output address select bit (A0=1)
	movwf	PORT_BUS_O		; write register address to data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	bcf	PIN_USB_A0		; output data select bit (A0=0)
	movff	A1_REG,PORT_BUS_O	; write data
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	bsf	PIN_USB_nCS		; clear chip select
	setf	PORT_BUS_T		; set data bus port back to input
; enable interrupts
;!!!
; done
	return

;-------------------------------------------------------------------------------------------
; write a usb chip register multiple times
; input:	W		register address (0..63)
;		FSR0		pointer to data
;		A1_REG		number of bytes
; changes:	A1_REG, FSR0
;-------------------------------------------------------------------------------------------
USB_WRITE_REG_EX:
; disable interrupts
;!!!
; write register
	bsf	PIN_USB_A0		; output address select bit (A0=1)
	movwf	PORT_BUS_O		; write register address to data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	bcf	PIN_USB_A0		; output data select bit (A0=0)
USB_WRITE_REG_EX_LOOP:
	movff	POSTINC0,PORT_BUS_O	; write data
	bcf	PIN_USB_nWR		; set write
	bsf	PIN_USB_nWR		; clear write
	decfsz	A1_REG			; loop
	bra	USB_WRITE_REG_EX_LOOP
	bsf	PIN_USB_nCS		; clear chip select
	setf	PORT_BUS_T		; set data bus port back to input
; enable interrupts
;!!!
; done
	return

;-------------------------------------------------------------------------------------------
; usb task
; - called periodically
; changes:	PRODH, PRODL, A1_REG, B1_REG, C1_REG, TBLPTR, TABLAT, WREG
;-------------------------------------------------------------------------------------------
USB_TASK:
; check if usb is initialized
	btfsc	USB_IS_INIT		; if usb is initialized ...
	bra	USB_PROCESS		; ... jump to processing
	btfsc	USB_IS_PRE_INIT		; if usb is pre-initialized ...
	bra	USB_INIT2		; ... jump to second part of initialization
	bra	USB_INIT1		; jump to first part of initialization

;-------------------------------------------------------------------------------------------
; usb initial state - 1st part
;-------------------------------------------------------------------------------------------
USB_INIT1:
; setup usb chip
	movlw	b'00000100'		; 3.3 volts on, no node attach, no interrupt output
	movwf	A1_REG
	movlw	USB_MCNTRL
	call	USB_WRITE_REG
	movlw	b'10001011'		; clock output off, divisor to 4 Mhz
	movwf	A1_REG
	movlw	USB_CCONF
	call	USB_WRITE_REG
; now, usb is pre-initialized
	bsf	USB_IS_PRE_INIT
; done
	return

;-------------------------------------------------------------------------------------------
; usb initial state - 2nd part
;-------------------------------------------------------------------------------------------
USB_INIT2:
; go on with initializing usb chip
	movlw	b'00001100'		; 3.3 volts still on, but node attach on, no interrupt output
	movwf	A1_REG
	movlw	USB_MCNTRL
	call	USB_WRITE_REG
; now, usb is initialized
	bsf	USB_IS_INIT
; enable NAK events
	movlw	b'11111111'		; NAKMSK: NAK_*=1
	movwf	A1_REG
	movlw	USB_NAKMSK
	call	USB_WRITE_REG
; read ALTEV to clear node reset bit
	movlw	USB_ALTEV
	call	USB_READ_REG
; done - continue with node reset
	bra	USB_NODE_RESET

;-------------------------------------------------------------------------------------------
; usb node reset processing
;-------------------------------------------------------------------------------------------
USB_NODE_RESET:
; set node functional state to node reset
	movlw	b'00000000'		; NFSR: NFS=NodeReset
	movwf	A1_REG
	movlw	USB_NFSR
	call	USB_WRITE_REG
; done - continue with start
	bra	USB_START

;-------------------------------------------------------------------------------------------
; usb start (after initialization or node reset)
;-------------------------------------------------------------------------------------------
USB_START:
; set node address to 0
	movlw	b'10000000'		; FAR: AD_EN=1, AD=0
	movwf	A1_REG
	movlw	USB_FAR
	call	USB_WRITE_REG
; initialize enpoint 0
	movlw	b'00000000'		; EPC0: STALL=0, DEF=0, EP=0
	movwf	A1_REG
	movlw	USB_EPC0
	call	USB_WRITE_REG
	movlw	b'00001000'		; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG			; FLUSH setzen und TX_EN rcksetzen
	movlw	USB_TXC0
	call	USB_WRITE_REG
; initialize variables
	bcf	USB_0_TOGGLE		; expect TOGGLE=0 for EP0
	bcf	USB_S0_SEND		; not sending a message on endpoint 0
	bcf	USB_S0_PACKET		; not sending a packet on endpoint 0
	clrf	USB_R0_CNT		; nothing received so far
	bcf	USB_CONF		; we are not configured
	call	USB_CONF_OFF
	bcf	BIT_USB_PWR_OK		; power may not be taken from usb
; allow reception
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
; set node functional state to operational
	movlw	b'00000010'		; NFSR: NFS=NodeOperational
	movwf	A1_REG
	movlw	USB_NFSR
	call	USB_WRITE_REG
; done
	return

;-------------------------------------------------------------------------------------------
; usb processing
;-------------------------------------------------------------------------------------------
USB_PROCESS:
; check for node reset
	movlw	USB_ALTEV		; read ALTEV
	call	USB_READ_REG
	btfsc	A1_REG,6		; ALTEV: RESET=1 ===> node reset was detected
	bra	USB_NODE_RESET

; sending on EP0?
	btfsc	USB_S0_PACKET		; sending a packet
	bra	USB_CONFIG_SEND_PACKET
	btfsc	USB_S0_SEND		; data to send is available
	bra	USB_CONFIG_SEND

; receive messages
; get receive status
	movlw	USB_RXS0		; read RXS0
	call	USB_READ_REG
	movf	A1_REG,W
	movwf	C1_REG			; remember RXS0 in C1_REG
	andlw	b'00001111'		; RXS0: RCOUNT - put number of received bytes into B1_REG
	movwf	B1_REG
	bz	USB_CONFIG_RECV_READ_DONE	; no bytes to read
; get recevied bytes
USB_CONFIG_RECV_READ:
	movf	USB_R0_CNT,W		; get offset into receive buffer
	andlw	b'00000111'		;   - use only last 3 bits because buffer length is 8
					;   - longest expected message is 8 bytes in length
	lfsr	FSR0,USB_R0_BUF		; get address within buffer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	USB_RXD0		; RXD0 - read a data byte from fifo
	call	USB_READ_REG
	movff	A1_REG,INDF0
	incf	USB_R0_CNT		; count received bytes
	decfsz	B1_REG			; count down bytes in fifo
	bra	USB_CONFIG_RECV_READ
USB_CONFIG_RECV_READ_DONE:
; check if message complete
	btfss	C1_REG,4		; if packet was not last one (RXS0: LAST=0) ...
	bra	USB_CONFIG_RECV_END	; ... jump to end of reception
; check if message is a valid setup message
	bcf	USB_0_TOGGLE		; store inverted toggle bit (used later for sending)
	btfss	C1_REG,5		; (RXS0: TOGGLE)
	bsf	USB_0_TOGGLE
	btfss	C1_REG,6		; if not a setup message (RXS0: SETUP=0) ...
	bra	USB_CONFIG_RECV_DONE	; ... ignore this message
	movf	USB_R0_CNT,W		; if not 8 bytes in length ...
	xorlw	8
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ	; ... treat this message as unknown request
; interpret setup message and prepare to send answer
	movlw	0x00			; bRequest=0x00?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_00
	movlw	0x01			; bRequest=0x01?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_01
	movlw	0x05			; bRequest=0x05?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_05
	movlw	0x06			; bRequest=0x06?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_06
	movlw	0x08			; bRequest=0x08?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_08
	movlw	0x09			; bRequest=0x09?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_09
	movlw	0x0A			; bRequest=0x0A?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_0A
	movlw	0x0B			; bRequest=0x0B?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_0B
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x00
USB_CONFIG_REQ_00:
	movlw	0x80			; bmRequestType=0x80?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_DEVICE
	movlw	0x81			; bmRequestType=0x81?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_INTERFACE
	movlw	0x82			; bmRequestType=0x82?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_ENDPOINT
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x01
USB_CONFIG_REQ_01:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_DEVICE
	movlw	0x01			; bmRequestType=0x01?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_INTERFACE
	movlw	0x02			; bmRequestType=0x02?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_ENDPOINT
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x05
USB_CONFIG_REQ_05:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_ADDRESS
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x06
USB_CONFIG_REQ_06:
	movlw	0x00			; bmRequestType=0x00/0x80?
	xorwf	USB_BM_REQ_TYPE,W
	andlw	0x7F
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_DESCR
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x08
USB_CONFIG_REQ_08:
	movlw	0x80			; bmRequestType=0x80?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_CONFIG
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x09
USB_CONFIG_REQ_09:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_CONFIG
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x0A
USB_CONFIG_REQ_0A:
	movlw	0x81			; bmRequestType=0x81?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_INTERFACE
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x0B
USB_CONFIG_REQ_0B:
	movlw	0x01			; bmRequestType=0x01?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_INTERFACE
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get status device/interface/endpoint"
USB_CONFIG_GET_STATUS_DEVICE:
USB_CONFIG_GET_STATUS_INTERFACE:
USB_CONFIG_GET_STATUS_ENDPOINT:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wIndex=0x0000, wLength=0x0002?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_INDEX+0,W
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x02
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_00_00	; send 0x00, 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get status device/interface/endpoint"
USB_CONFIG_CLEAR_FEAT_DEVICE:
	movwf	USB_W_INDEX+0,W		; wIndex=x0000?
	iorwf	USB_W_INDEX+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
USB_CONFIG_CLEAR_FEAT_INTERFACE:
USB_CONFIG_CLEAR_FEAT_ENDPOINT:
	movwf	USB_W_LENGTH+0,W	; wLength=x0000?
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set address"
USB_CONFIG_SET_ADDRESS:
	movlw	b'01000000'		; EPC0: STALL=0, DEF=1, EP=0
	movwf	A1_REG
	movlw	USB_EPC0
	call	USB_WRITE_REG
	movlw	b'10000000'		; FAR: AD_EN=1, AD=address
	iorwf	USB_W_VALUE+0,W
	movwf	A1_REG
	movlw	USB_FAR		
	call	USB_WRITE_REG
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get descriptor"
USB_CONFIG_GET_DESCR:
	movlw	0x01			; wValue=0x01__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_DEV
	movlw	0x02			; wValue=0x02__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_CONF
	movlw	0x03			; wValue=0x03__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_STR
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get device descriptor"
USB_CONFIG_GET_DESCR_DEV:
	movlw	low USB_DESCR_DEV	; send device desriptor
	movwf	TBLPTRL
	movlw	high USB_DESCR_DEV
	movwf	TBLPTRH
	movlw	upper USB_DESCR_DEV
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get configuration descriptor"
USB_CONFIG_GET_DESCR_CONF:
	movlw	low USB_DESCR_CONF	; send configuration desriptor (with subdescriptors)
	movwf	TBLPTRL
	movlw	high USB_DESCR_CONF
	movwf	TBLPTRH
	movlw	upper USB_DESCR_CONF
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor"
USB_CONFIG_GET_DESCR_STR:
	movlw	0x00			; wValue=0x__00?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR0
	movlw	0x01			; wValue=0x__01?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR1
	movlw	0x02			; wValue=0x__02?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR2
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get string descriptor 0"
USB_CONFIG_GET_DESCR_STR0:
	movlw	low USB_DESCR_STR0	; send string desriptor 0
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR0
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR0
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor 1"
USB_CONFIG_GET_DESCR_STR1:
	movlw	low USB_DESCR_STR1	; send string desriptor 1
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR1
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR1
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor 2"
USB_CONFIG_GET_DESCR_STR2:
	movlw	low USB_DESCR_STR2	; send string desriptor 2
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR2
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR2
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get configuration"
USB_CONFIG_GET_CONFIG:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wIndex=0x0000, wLength=0x0001?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_INDEX+0,W
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x01
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	btfsc	USB_CONF		; are we configured?
	bra	USB_CONFIG_GET_CONFIG_YES
	movlw	low USB_ANSWER_00	; send 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
USB_CONFIG_GET_CONFIG_YES:
	movlw	low USB_ANSWER_01	; send 0x01
	movwf	TBLPTRL
	movlw	high USB_ANSWER_01
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_01
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set configuration"
USB_CONFIG_SET_CONFIG:
	movwf	USB_W_INDEX+0,W		; wIndex=0x0000, wLength=0x0001?
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movf	USB_W_VALUE,W		; configuration or deconfiguration?
	bnz	USB_CONFIG_SET_CONFIG_YES
	bcf	USB_CONF		; deconfiguration
	call	USB_CONF_OFF
	bcf	BIT_USB_PWR_OK		; power may not be taken from usb
	bra	USB_CONFIG_SET_CONFIG_ANSWER
USB_CONFIG_SET_CONFIG_YES:
	bsf	USB_CONF		; configuration
	call	USB_CONF_ON
	bsf	BIT_USB_PWR_OK		; power may be taken from usb
	bcf	USB_S1_TOGGLE		; TOGGLE=0 for first packet sent on endpoint 1
	bcf	USB_S1_ACT		; not sending a message on endpoint 1
	bcf	USB_S2_TOGGLE		; TOGGLE=0 for first packet sent on endpoint 2
	bcf	USB_S2_ACT		; not sending a message on endpoint 2
	bcf	USB_S3_TOGGLE		; TOGGLE=0 for first packet sent on endpoint 3
	bcf	USB_S3_ACT		; not sending a message on endpoint 3
	movlw	1			; send first packet using endpoint 1
	movwf	USB_Sx_EP
	clrf	USB_R1_CNT		; no packet received yet
	clrf	USB_R2_CNT
	clrf	USB_R3_CNT
	clrf	USB_Sx_CNT		; no packet to send yet
	clrf	SEQ_NO_RECV		; clear sequence numbers
	clrf	SEQ_NO_SEND
USB_CONFIG_SET_CONFIG_ANSWER:
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get interface"
USB_CONFIG_GET_INTERFACE:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wLength=0x0001?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x01
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_00	; send 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set interface"
USB_CONFIG_SET_INTERFACE:
	movwf	USB_W_LENGTH+0,W	; wLength=0x0001?
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
;unknown request
USB_CONFIG_UNKNOWN_REQ:
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; received message is processed
USB_CONFIG_RECV_DONE:
; enable reception of new message on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
; end of reception branch of configuration stuff
USB_CONFIG_RECV_END:
	bra	USB_CONFIG_END

; start sending answer at TBLPTR
USB_CONFIG_ANSWER:
	setf	A1_REG			; any length is allowed
; start sending answer at TBLPTR with max length A1_REG
USB_CONFIG_ANSWER_LEN:
	movlw	0x80			; configure for code read
	movwf	EECON1
	tblrd*+				; read length of table
	movff	TABLAT,USB_S0_LEN
	tblrd*+				; increment pointer
	movf	A1_REG,W		; max. length from A1_REG
	cpfslt	USB_S0_LEN
	movwf	USB_S0_LEN
	movff	TBLPTRL,USB_S0_PTR+0	; save pointer
	movff	TBLPTRH,USB_S0_PTR+1
	movff	TBLPTRU,USB_S0_PTR+2
	bsf	USB_S0_SEND		; data to send is now available
	bcf	USB_S0_PACKET		; no packet is being sent at the moment
					; - packet transmission is aborted if active
	bra	USB_CONFIG_SEND

; sending a packet
USB_CONFIG_SEND_PACKET:
; get send status
	movlw	USB_TXS0		; read TXS0
	call	USB_READ_REG		; and store its contents in A1_REG
; check if packet was sent
	btfss	A1_REG,5		; jump if packet was not send (TXS0: TX_DONE=0)
	bra	USB_CONFIG_SEND_END
; check if packet was successfully sent
	btfsc	A1_REG,6		; jump if packet was successfully sent (TXS0: ACK_STAT=1)
	bra	USB_CONFIG_SEND_OK
; restore old state, packet needs to be sent again
	movff	USB_S0_LEN_OLD,USB_S0_LEN
	movff	USB_S0_PTR_OLD+0,USB_S0_PTR+0
	movff	USB_S0_PTR_OLD+1,USB_S0_PTR+1
	movff	USB_S0_PTR_OLD+2,USB_S0_PTR+2
	bsf	USB_S0_SEND		; there is data to be sent
USB_CONFIG_SEND_OK:
	bcf	USB_S0_PACKET		; no packet is being sent
	btg	USB_0_TOGGLE		; invert toggle bit
; check if there is more data to send
	btfsc	USB_S0_SEND		; jump if more data to send is available
	bra	USB_CONFIG_SEND
; no more data to send available
; flush FIFO and enable reception on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	bra	USB_CONFIG_SEND_END

; data to send is available - load the next packet to send
; if we get here, USB_S0_SEND is always set and data to send is available
USB_CONFIG_SEND:
; disable reception and clear buffer on EP0
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00001000'		; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
; save state for the case that packet fails and needs to be sent again
	movff	USB_S0_LEN,USB_S0_LEN_OLD
	movff	USB_S0_PTR+0,USB_S0_PTR_OLD+0
	movff	USB_S0_PTR+1,USB_S0_PTR_OLD+1
	movff	USB_S0_PTR+2,USB_S0_PTR_OLD+2
; check if to send 8 bytes or more
	movlw	8			; subtract 8 from length
	subwf	USB_S0_LEN,W		; and store result in WREG
	btfss	STATUS,C
	bra	USB_CONFIG_SEND_LAST_PACKET	; jump if packet is less than 8 bytes
	movwf	USB_S0_LEN		; save new length (now 8 bytes less)
	movlw	8			; 8 bytes to load into fifo
	movwf	B1_REG
	bra	USB_CONFIG_SEND_LOAD_FIFO
; prepare for last packet
USB_CONFIG_SEND_LAST_PACKET:
	movff	USB_S0_LEN,B1_REG	; remember number of bytes to load into fifo
	clrf	USB_S0_LEN		; entire message has been sent
	bcf	USB_S0_SEND
	movf	B1_REG,W		; check if empty packet
	bz	USB_CONFIG_SEND_START	; nothing needs to be loaded into fifo for an empty packet
; load a packet data into fifo
USB_CONFIG_SEND_LOAD_FIFO:
	movff	USB_S0_PTR+0,TBLPTRL	; get pointer to rest of packet
	movff	USB_S0_PTR+1,TBLPTRH
	movff	USB_S0_PTR+2,TBLPTRU
	movlw	0x80			; configure for code read
	movwf	EECON1
USB_CONFIG_SEND_LOAD_FIFO_LOOP:
	tblrd*+				; get next byte
	movff	TABLAT,A1_REG		; write byte into fifo
	movlw	USB_TXD0
	call	USB_WRITE_REG
	decfsz	B1_REG			; count down bytes to send
	bra	USB_CONFIG_SEND_LOAD_FIFO_LOOP
	movff	TBLPTRL,USB_S0_PTR+0	; save new pointer to rest of packet
	movff	TBLPTRH,USB_S0_PTR+1
	movff	TBLPTRU,USB_S0_PTR+2
; start sending the packet
USB_CONFIG_SEND_START:
	bsf	USB_S0_PACKET		; packet is being sent now
	movlw	b'00000001'		; TXC0: IGN_IN=0, FLUSH=0, TOGGLE=0, TX_EN=1
	btfsc	USB_0_TOGGLE		; write right toggle bit
	iorlw	b'00000100'
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
	movlw	USB_NAKEV		; read NAKEV to clear NAK events
	call	USB_READ_REG		; (it might be possible that there is an old NAK event pending)
; end of configuration send stuff
USB_CONFIG_SEND_END:

; check for EP0 BULK OUT NAK events
; - i.e. if host tries to send and we are not listening
	movlw	USB_NAKEV		; read NAKEV
	call	USB_READ_REG		; and store its contents in A1_REG
	btfss	A1_REG,4		; check if NAK bit for BULK OUT on EP0 is set
	bra	USB_CONFIG_NAK_END	; not set
; clear send bits for EP0
	bcf	USB_S0_SEND
	bcf	USB_S0_PACKET
; cancel transmission on EP0
	movlw	b'00001000'		; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
; clear FIFO and enable reception on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	bra	USB_CONFIG_SEND_END
; end of configuration NAK stuff
USB_CONFIG_NAK_END:

; end of configuration stuff
USB_CONFIG_END:

; return if not configured
	btfss	USB_CONF
	return

; check if receive buffer 1 is empty
	movf	USB_R1_CNT,W
	bnz	USB_R1_END
; receive packets on data endpoint 1
	movlw	USB_RXS1		; get receive state from RXS1
	call	USB_READ_REG
	btfss	A1_REG,7		; jump if no error (RX_ERR=0)
	bra	USB_R1_NO_ERR
	movlw	b'00001100'		; RXC1: FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1 
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC1: FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1 
	call	USB_WRITE_REG
	bra	USB_R1_END
USB_R1_NO_ERR:
	btfss	A1_REG,4		; jump to end if packet not complete (RX_LAST=0)
	bra	USB_R1_END
	clrf	USB_R1_CNT		; clear number of received bytes
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfsc	STATUS,Z		; jump if empty packet
	bra	USB_R1_GO_ON
; read FIFO
	lfsr	FSR0,USB_R1_BUF		; get address of receive buffer
USB_R1_LOOP:
	movf	B1_REG,W		; add up number of bytes
	addwf	USB_R1_CNT
	movwf	A1_REG			; read bytes from FIFO
	movlw	USB_RXD1
	call 	USB_READ_REG_EX
	movlw	USB_RXS1		; get receive status
	call	USB_READ_REG
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfss	STATUS,Z
	bra	USB_R1_LOOP
; reenable reception on endpoint 1
USB_R1_GO_ON:
	movlw	b'00001100' 		; RXC1: RFWL=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00000101' 		; RXC1: RFWL=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
USB_R1_END:

; send packets on data endpoint 1
	btfss	USB_S1_ACT		; jump if not currently sending
	bra	USB_S1_NEW		
	movlw	USB_TXS1		; get transmit state from TXS1
	call	USB_READ_REG
	btfss	A1_REG,5		; jump if TX_DONE not set
	bra	USB_S1_END
	btfsc	A1_REG,6		; jump if ACK_STAT set
	bra	USB_S1_ACK
; transmission failed (no ACK), repeat it
	movlw	b'00010000'	 	; TXC1: RFF=1
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	movlw	b'00000011'		; TXC1: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S1_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	bra	USB_S1_END
; transmission succeeded
USB_S1_ACK:
	bcf	USB_S1_ACT		; no more sending
	btg	USB_S1_TOGGLE		; invert toggle bit
; start a new transmission
USB_S1_NEW:
	movf	USB_Sx_CNT		; jump if nothing to send
	bz	USB_S1_END
	movlw	1			; jump if it is not EP1's turn
	cpfseq	USB_Sx_EP
	bra	USB_S1_END
	movlw	b'00001000'		; TXC1: FLUSH=1
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	movff	SEQ_NO_SEND,A1_REG	; load sequence number into FIFO (TXD1)
	movlw	USB_TXD1
	call	USB_WRITE_REG
	incf	SEQ_NO_SEND		; increment sequence number	
	lfsr	FSR0,USB_Sx_BUF		; load data into FIFO (TXD1)
	movff	USB_Sx_CNT,A1_REG
	movlw	USB_TXD1
	call	USB_WRITE_REG_EX
	movlw	b'00000011'		; TXC1: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S1_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	bsf	USB_S1_ACT		; now sending
	clrf	USB_Sx_CNT		; now there is nothing to send
	movlw	2			; use EP 2 next time
	movwf	USB_Sx_EP
USB_S1_END:

; check if receive buffer 2 is empty
	movf	USB_R2_CNT,W
	bnz	USB_R2_END
; receive packets on data endpoint 2
	movlw	USB_RXS2		; get receive state from RXS2
	call	USB_READ_REG
	btfss	A1_REG,7		; jump if no error (RX_ERR=0)
	bra	USB_R2_NO_ERR
	movlw	b'00001100'		; RXC2: FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC2 
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC2: FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC2 
	call	USB_WRITE_REG
	bra	USB_R2_END
USB_R2_NO_ERR:
	btfss	A1_REG,4		; jump to end if packet not complete (RX_LAST=0)
	bra	USB_R2_END
	clrf	USB_R2_CNT		; clear number of received bytes
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfsc	STATUS,Z		; jump if empty packet
	bra	USB_R2_GO_ON
; read FIFO
	lfsr	FSR0,USB_R2_BUF		; get address of receive buffer
USB_R2_LOOP:
	movf	B1_REG,W		; add up number of bytes
	addwf	USB_R2_CNT
	movwf	A1_REG			; read bytes from FIFO
	movlw	USB_RXD2
	call 	USB_READ_REG_EX
	movlw	USB_RXS2		; get receive status
	call	USB_READ_REG
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfss	STATUS,Z
	bra	USB_R2_LOOP
; reenable reception on endpoint 2
USB_R2_GO_ON:
	movlw	b'00001100' 		; RXC2: RFWL=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC2
	call	USB_WRITE_REG
	movlw	b'00000101' 		; RXC2: RFWL=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC2
	call	USB_WRITE_REG
USB_R2_END:

; send packets on data endpoint 2
	btfss	USB_S2_ACT		; jump if not currently sending
	bra	USB_S2_NEW		
	movlw	USB_TXS2		; get transmit state from TXS2
	call	USB_READ_REG
	btfss	A1_REG,5		; jump if TX_DONE not set
	bra	USB_S2_END
	btfsc	A1_REG,6		; jump if ACK_STAT set
	bra	USB_S2_ACK
; transmission failed (no ACK), repeat it
	movlw	b'00010000'	 	; TXC2: RFF=1
	movwf	A1_REG
	movlw	USB_TXC2
	call	USB_WRITE_REG
	movlw	b'00000011'		; TXC2: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S2_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC2
	call	USB_WRITE_REG
	bra	USB_S2_END
; transmission succeeded
USB_S2_ACK:
	bcf	USB_S2_ACT		; no more sending
	btg	USB_S2_TOGGLE		; invert toggle bit
; start a new transmission
USB_S2_NEW:
	movf	USB_Sx_CNT		; jump if nothing to send
	bz	USB_S2_END
	movlw	2			; jump if it is not EP2's turn
	cpfseq	USB_Sx_EP
	bra	USB_S2_END
	movlw	b'00001000'		; TXC2: FLUSH=1
	movwf	A1_REG
	movlw	USB_TXC2
	call	USB_WRITE_REG
	movff	SEQ_NO_SEND,A1_REG	; load sequence number into FIFO (TXD2)
	movlw	USB_TXD2
	call	USB_WRITE_REG
	incf	SEQ_NO_SEND		; increment sequence number	
	lfsr	FSR0,USB_Sx_BUF		; load data into FIFO (TXD2)
	movff	USB_Sx_CNT,A1_REG
	movlw	USB_TXD2
	call	USB_WRITE_REG_EX
	movlw	b'00000011'		; TXC2: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S2_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC2
	call	USB_WRITE_REG
	bsf	USB_S2_ACT		; now sending
	clrf	USB_Sx_CNT		; now there is nothing to send
	movlw	3			; use EP 3 next time
	movwf	USB_Sx_EP
USB_S2_END:

; check if receive buffer 3 is empty
	movf	USB_R3_CNT,W
	bnz	USB_R3_END
; receive packets on data endpoint 3
	movlw	USB_RXS3		; get receive state from RXS3
	call	USB_READ_REG
	btfss	A1_REG,7		; jump if no error (RX_ERR=0)
	bra	USB_R3_NO_ERR
	movlw	b'00001100'		; RXC3: FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC3 
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC3: FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC3 
	call	USB_WRITE_REG
	bra	USB_R3_END
USB_R3_NO_ERR:
	btfss	A1_REG,4		; jump to end if packet not complete (RX_LAST=0)
	bra	USB_R3_END
	clrf	USB_R3_CNT		; clear number of received bytes
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfsc	STATUS,Z		; jump if empty packet
	bra	USB_R3_GO_ON
; read FIFO
	lfsr	FSR0,USB_R3_BUF		; get address of receive buffer
USB_R3_LOOP:
	movf	B1_REG,W		; add up number of bytes
	addwf	USB_R3_CNT
	movwf	A1_REG			; read bytes from FIFO
	movlw	USB_RXD3
	call 	USB_READ_REG_EX
	movlw	USB_RXS3		; get receive status
	call	USB_READ_REG
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfss	STATUS,Z
	bra	USB_R3_LOOP
; reenable reception on endpoint 3
USB_R3_GO_ON:
	movlw	b'00001100' 		; RXC3: RFWL=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC3
	call	USB_WRITE_REG
	movlw	b'00000101' 		; RXC3: RFWL=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC3
	call	USB_WRITE_REG
USB_R3_END:

; send packets on data endpoint 3
	btfss	USB_S3_ACT		; jump if not currently sending
	bra	USB_S3_NEW		
	movlw	USB_TXS3		; get transmit state from TXS3
	call	USB_READ_REG
	btfss	A1_REG,5		; jump if TX_DONE not set
	bra	USB_S3_END
	btfsc	A1_REG,6		; jump if ACK_STAT set
	bra	USB_S3_ACK
; transmission failed (no ACK), repeat it
	movlw	b'00010000'	 	; TXC3: RFF=1
	movwf	A1_REG
	movlw	USB_TXC3
	call	USB_WRITE_REG
	movlw	b'00000011'		; TXC3: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S3_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC3
	call	USB_WRITE_REG
	bra	USB_S3_END
; transmission succeeded
USB_S3_ACK:
	bcf	USB_S3_ACT		; no more sending
	btg	USB_S3_TOGGLE		; invert toggle bit
; start a new transmission
USB_S3_NEW:
	movf	USB_Sx_CNT		; jump if nothing to send
	bz	USB_S3_END
	movlw	3			; jump if it is not EP3's turn
	cpfseq	USB_Sx_EP
	bra	USB_S3_END
	movlw	b'00001000'		; TXC3: FLUSH=1
	movwf	A1_REG
	movlw	USB_TXC3
	call	USB_WRITE_REG
	movff	SEQ_NO_SEND,A1_REG	; load sequence number into FIFO (TXD3)
	movlw	USB_TXD3
	call	USB_WRITE_REG
	incf	SEQ_NO_SEND		; increment sequence number	
	lfsr	FSR0,USB_Sx_BUF		; load data into FIFO (TXD3)
	movff	USB_Sx_CNT,A1_REG
	movlw	USB_TXD3
	call	USB_WRITE_REG_EX
	movlw	b'00000011'		; TXC3: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S3_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC3
	call	USB_WRITE_REG
	bsf	USB_S3_ACT		; now sending
	clrf	USB_Sx_CNT		; now there is nothing to send
	movlw	1			; use EP 1 next time
	movwf	USB_Sx_EP
USB_S3_END:

; done
	return

;-------------------------------------------------------------------------------------------
; turn on usb data channels
;-------------------------------------------------------------------------------------------
USB_CONF_ON:
	movlw	b'00001000'		; TXC1: IGN_ISOMSK=0, TFWL=0, RRF=0, FLUSH=1, TOGGLE=0, LAST=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	movlw	b'00010001'		; EPC1: STALL=0, ISO=0, EP_EN=1, ADDR=1
	movwf	A1_REG
	movlw	USB_EPC1
	call	USB_WRITE_REG
	movlw	b'00001100'		; RXC1: RFWL=0, IGN_ISO=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC1: RFWL=0, IGN_ISO=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00010001'		; EPC2: STALL=0, ISO=0, EP_EN=1, ADDR=1
	movwf	A1_REG
	movlw	USB_EPC2
	call	USB_WRITE_REG
	movlw	b'00001000'		; TXC2: IGN_ISOMSK=0, TFWL=0, RRF=0, FLUSH=1, TOGGLE=0, LAST=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC2
	call	USB_WRITE_REG
	movlw	b'00010010'		; EPC3: STALL=0, ISO=0, EP_EN=1, ADDR=2
	movwf	A1_REG
	movlw	USB_EPC3
	call	USB_WRITE_REG
	movlw	b'00001100'		; RXC2: RFWL=0, IGN_ISO=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC2
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC2: RFWL=0, IGN_ISO=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC2
	call	USB_WRITE_REG
	movlw	b'00010010'		; EPC4: STALL=0, ISO=0, EP_EN=1, ADDR=2
	movwf	A1_REG
	movlw	USB_EPC4
	call	USB_WRITE_REG
	movlw	b'00001000'		; TXC3: IGN_ISOMSK=0, TFWL=0, RRF=0, FLUSH=1, TOGGLE=0, LAST=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC3
	call	USB_WRITE_REG
	movlw	b'00010011'		; EPC5: STALL=0, ISO=0, EP_EN=1, ADDR=3
	movwf	A1_REG
	movlw	USB_EPC5
	call	USB_WRITE_REG
	movlw	b'00001100'		; RXC3: RFWL=0, IGN_ISO=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC3
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC3: RFWL=0, IGN_ISO=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC3
	call	USB_WRITE_REG
	movlw	b'00010011'		; EPC6: STALL=0, ISO=0, EP_EN=1, ADDR=3
	movwf	A1_REG
	movlw	USB_EPC6
	call	USB_WRITE_REG
	return

;-------------------------------------------------------------------------------------------
; turn off usb data channels
;-------------------------------------------------------------------------------------------
USB_CONF_OFF:
	movlw	b'00000000'		; EPC1: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC1
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC2: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC2
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC3: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC3
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC4: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC4
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC5: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC5
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC6: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC6
	call	USB_WRITE_REG
	return

;-------------------------------------------------------------------------------------------
; standard usb answers
;-------------------------------------------------------------------------------------------
USB_ANSWER_EMPTY:
	DB	low (USB_ANSWER_EMPTY_END - USB_ANSWER_EMPTY_START)
USB_ANSWER_EMPTY_START:
USB_ANSWER_EMPTY_END:

USB_ANSWER_00:
	DB	low (USB_ANSWER_00_END - USB_ANSWER_00_START)
USB_ANSWER_00_START:
	DB	0x00
USB_ANSWER_00_END:

USB_ANSWER_01:
	DB	low (USB_ANSWER_01_END - USB_ANSWER_01_START)
USB_ANSWER_01_START:
	DB	0x01
USB_ANSWER_01_END:


USB_ANSWER_00_00:
	DB	low (USB_ANSWER_00_00_END - USB_ANSWER_00_00_START)
USB_ANSWER_00_00_START:
	DB	0x00,0x00
USB_ANSWER_00_00_END:

;-------------------------------------------------------------------------------------------
; usb descriptors
;-------------------------------------------------------------------------------------------

USB_DESCR_DEV:
	DB	low (USB_DESCR_DEV_END - USB_DESCR_DEV_START)
USB_DESCR_DEV_START:
	DB	0x12,0x01	; bLength: size of descriptor, bType: device descriptor
	DB	0x10,0x01	; bcdUSB: USB 1.1.0
	DB	0x00,0x00	; bDeviceClass: no class, bDeviceSubClass: no subclass
	DB	0x00,0x08	; bDeviceProtocol: no protocol,bMaxPacketSize: 8 bytes per packet
	DB	0x66,0x66	; idVendor: 0x6666 prototype vendor id
	DB	0x53,0x33	; idProduct: 0x3353 '3S' for Stefan Schuermans Software
	DB	0x01,0x03	; bcdDevice: 0x0300=HackerPort (0x03__), Revision 1 (0x__01)
	DB	0x01,0x02	; iManufacturer: index of manufacturer string, iProduct: index of product string
	DB	0x00,0x01	; iSerial: no serial number string,
USB_DESCR_DEV_END:		; bNumConfigurations: we only have one possible configuration


USB_DESCR_CONF:
	DB	low (USB_DESCR_CONF_END - USB_DESCR_CONF_START)
USB_DESCR_CONF_START:
; ConfigurationDescriptor
	DB	0x09,0x02	; bLength: size of descriptor
				; bType: configuration descriptor
	DB	0x3C,0x00	; wTotalLength: total size of entire configuration
	DB	0x01,0x01	; bNumInterfaces: one interface
				; bConfigurationValue: configuration identifier (our configuration is called 0x01)
	DB	0x00,0x80	; iConfiguration: no configuration string
				; bmAttributes: not self powered, no remote wakeup
	DB	0xFA,0x09	; bMaxPower: power consumption max. 500mA
; InterfaceDescriptor		; bLength: size of descriptor
	DB	0x04,0x00	; bType: interface descriptor
				; bInterfaceNumber: this is the first interface
	DB	0x00,0x06	; bAlternateSetting: no alternate settings
				; bNumEndpoints: 6 endpoints (without endpoint 0)
	DB	0xFF,0xFF	; bInterfaceClass: we don't have a registered class code
				; bInterfaceSubClass: we don't have a registered subclass code
	DB	0xFF,0x00	; bInterfaceProtocol: we don't have a registered protocol code
				; iInterface: no interface string
; EndpointDescriptor1
	DB	0x07,0x05	; bLength: size of descriptor
				; bType: endpoint descriptor
	DB	0x01,0x02	; bEndpointAddress: endpoint 1, out
				; bmAttributes: bulk
	DB	0x40,0x00	; wMaxPacketSize: max. packet size 64
				; highbyte max. packet size
	DB	0x00,0x07	; bInterval: ignored for bulk
; EndpointDescriptor2		; bLength: size of descriptor
	DB	0x05,0x81	; bType: endpoint descriptor
				; bEndpointAddress: endpoint 1, in
	DB	0x02,0x40	; bmAttributes: bulk
				; wMaxPacketSize: max. packet size 64
	DB	0x00,0x00	; highbyte max. packet size
				; bInterval: ignored for bulk
; EndpointDescriptor3
	DB	0x07,0x05	; bLength: size of descriptor
				; bType: endpoint descriptor
	DB	0x02,0x02	; bEndpointAddress: endpoint 2, out
				; bmAttributes: bulk
	DB	0x40,0x00	; wMaxPacketSize: max. packet size 64
				; highbyte max. packet size
	DB	0x00,0x07	; bInterval: ignored for bulk
; EndpointDescriptor4		; bLength: size of descriptor
	DB	0x05,0x82	; bType: endpoint descriptor
				; bEndpointAddress: endpoint 2, in
	DB	0x02,0x40	; bmAttributes: bulk
				; wMaxPacketSize: max. packet size 64
	DB	0x00,0x00	; highbyte max. packet size
				; bInterval: ignored for bulk
; EndpointDescriptor5
	DB	0x07,0x05	; bLength: size of descriptor
				; bType: endpoint descriptor
	DB	0x03,0x02	; bEndpointAddress: endpoint 3, out
				; bmAttributes: bulk
	DB	0x40,0x00	; wMaxPacketSize: max. packet size 64
				; highbyte max. packet size
	DB	0x00,0x07	; bInterval: ignored for bulk
; EndpointDescriptor6		; bLength: size of descriptor
	DB	0x05,0x83	; bType: endpoint descriptor
				; bEndpointAddress: endpoint 3, in
	DB	0x02,0x40	; bmAttributes: bulk
				; wMaxPacketSize: max. packet size 64
	DB	0x00,0x00	; highbyte max. packet size
				; bInterval: ignored for bulk
USB_DESCR_CONF_END:

USB_DESCR_STR0:
	DB	low (USB_DESCR_STR0_END - USB_DESCR_STR0_START)
USB_DESCR_STR0_START:
	DB	0x04,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DB	0x09,0x04	; wLangId: US English
USB_DESCR_STR0_END:

USB_DESCR_STR1:
	DB	low (USB_DESCR_STR1_END - USB_DESCR_STR1_START)
USB_DESCR_STR1_START:
	DB	0x36,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DW	'S'
	DW	't'
	DW	'e'
	DW	'f'
	DW	'a'
	DW	'n'
	DW	' '
	DW	'S'
	DW	'c'
	DW	'h'
	DW	'u'
	DW	'e'
	DW	'r'
	DW	'm'
	DW	'a'
	DW	'n'
	DW	's'
	DW	' '
	DW	'S'
	DW	'o'
	DW	'f'
	DW	't'
	DW	'w'
	DW	'a'
	DW	'r'
	DW	'e'
USB_DESCR_STR1_END:

USB_DESCR_STR2:
	DB	low (USB_DESCR_STR2_END - USB_DESCR_STR2_START)
USB_DESCR_STR2_START:
	DB	0x1E,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DW	'H'
	DW	'a'
	DW	'c'
	DW	'k'
	DW	'e'
	DW	'r'
	DW	'P'
	DW	'o'
	DW	'r'
	DW	't'
	DW	' '
	DW	'0'
	DW	'.'
	DW	'3'
USB_DESCR_STR2_END:



;===========================================================================================
; global variables
;===========================================================================================

	global		USB_R1_CNT, USB_R1_BUF
	global		USB_R2_CNT, USB_R2_BUF
	global		USB_R3_CNT, USB_R3_BUF
	global		USB_Sx_CNT, USB_Sx_BUF



;===========================================================================================
; global procedures
;===========================================================================================
	global		USB_INIT, USB_TASK



	end
